import logging
import ipaddress
from functools import wraps

from ratelimit.exceptions import Ratelimited

from .models import Rule
from .middleware import _add_rule as _middleware_add_rule


logger = logging.getLogger(__name__)


def blacklist_ratelimited(duration, block=True):
    def decorator(fn):
        @wraps(fn)
        def wrapper(request, *args, **kwargs):
            if request.limited:
                if user_duration and request.user.is_authenticated:
                    _create_user_rule(request, user_duration)

                elif ip_duration:
                    _create_ip_rule(request, ip_duration)

                else:
                    logger.warning('Unable to blacklist ratelimited client.')

                if block:
                    raise Ratelimited()

            return fn(request, *args, **kwargs)

        return wrapper

    if isinstance(duration, tuple):
        user_duration, ip_duration = duration
    else:
        user_duration = ip_duration = duration

    assert user_duration or ip_duration

    return decorator


def _create_user_rule(request, duration):
    user = request.user
    comments = _create_comments(request)
    rule = Rule(user=user, duration=duration, comments=comments)
    _add_rule(rule)


def _create_ip_rule(request, duration):
    addr = request.META['REMOTE_ADDR']
    ip = ipaddress.ip_address(addr)
    comments = _create_comments(request)
    rule = Rule(address=ip.compressed, duration=duration, comments=comments)
    _add_rule(rule)


def _create_comments(request):
    request_id = getattr(request, 'id', '')
    comments = 'Automatically blacklisted ratelimited client.\n' \
        f'Request ID: {request_id}\n' \
        f'Request line: {request.method} {request.get_full_path()}\n'
    return comments


def _add_rule(rule):
    rule.save()
    _middleware_add_rule(rule)



import ipaddress
import threading
from typing import Optional, Set

from django.core.exceptions import SuspiciousOperation

from .models import Rule


_blacklist: Optional[Set[Rule]] = None
_blacklist_lock = threading.Lock()


class Blacklisted(SuspiciousOperation):
    pass


def blacklist_middleware(get_response):
    def middleware(request):
        if _blacklist is None:
            _load_blacklist()

        _filter_client(request)

        return get_response(request)

    return middleware


def _filter_client(request):
    client_user = request.user
    client_user_id = client_user.id

    client_addr = request.META['REMOTE_ADDR']
    client_ip = ipaddress.ip_address(client_addr)

    inactive_rules = set()

    for rule in _blacklist:
        if rule.is_active():
            # no logging here, since the event will be logged by django.security

            rule_user_id = rule.user_id
            if rule_user_id is not None and client_user_id == rule_user_id:
                message = 'Blacklisted client user: %s' % client_user.username
                raise Blacklisted(message)

            rule_network = rule.get_network()
            if rule_network is not None and client_ip in rule_network:
                message = 'Blacklisted client address: %s' % client_addr
                raise Blacklisted(message)

        else:
            inactive_rules.add(rule)

        if inactive_rules:
            _remove_rules(inactive_rules)


def _load_blacklist():
    global _blacklist

    with _blacklist_lock:
        _blacklist = {rule for rule in Rule.objects.all() if rule.is_active()}


def _add_rule(rule):
    global _blacklist

    with _blacklist_lock:
        _blacklist = _blacklist | {rule}


def _remove_rules(rules):
    global _blacklist

    with _blacklist_lock:
        _blacklist = _blacklist - rules



import os

from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()


setup(
    name='django-blacklist',

    version='0.1.0',

    description='Blacklist users and hosts in Django. Automatically blacklist rate-limited clients.',
    long_description=README,
    long_description_content_type='text/markdown',

    url='https://github.com/vsemionov/django-blacklist',

    author='Victor Semionov',
    author_email='vsemionov@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security'
    ],

    keywords='django blacklist ratelimit firewall',

    packages=find_packages(),

    install_requires=[
        'Django'
    ]
)



# Generated by Django 2.2.2 on 2019-06-13 00:08

import datetime
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('address', models.GenericIPAddressField(blank=True, null=True)),
                ('prefixlen', models.PositiveIntegerField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(128)])),
                ('duration', models.DurationField(validators=[django.core.validators.MinValueValidator(datetime.timedelta(0))])),
                ('comments', models.TextField(blank=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



import ipaddress
from datetime import timedelta

from django.db import models
from django.core import validators
from django.utils.timezone import now
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError


class Rule(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    user = models.ForeignKey(User, null=True, blank=True, on_delete=models.CASCADE)

    address = models.GenericIPAddressField(null=True, blank=True)
    prefixlen = models.PositiveIntegerField(null=True, blank=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(128)])

    duration = models.DurationField(validators=[validators.MinValueValidator(timedelta(0))])

    comments = models.TextField(blank=True)

    def __str__(self):
        return str(self.id)

    def get_network(self):
        addr = self.address

        if addr:
            if self.prefixlen is not None:
                addr += f'/{self.prefixlen}'
            return ipaddress.ip_network(addr, strict=False)

        else:
            return None

    def get_expires(self):
        return self.created + self.duration

    def is_active(self):
        return now() < self.get_expires()

    def clean(self):
        if not self.address:
            # blank input produces empty string
            self.address = None

        if self.address is None and self.prefixlen is not None:
            raise ValidationError('Prefixlen without address.')

        try:
            self.get_network()

        except ValueError as ve:
            raise ValidationError(str(ve))

        if self.user is None and self.address is None:
            raise ValidationError('Neither user nor address provided.')

        if self.user is not None and self.address is not None:
            raise ValidationError('Both user and address provided.')



from django.contrib import admin

from .models import Rule


class RuleAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'address', 'prefixlen', 'created', 'duration', 'is_active')
    list_filter = ('created', 'duration')
    search_fields = ('user__username', 'address', 'comments')


admin.site.register(Rule, RuleAdmin)
